/*
 * Copyright (c) 1999.  Sun Microsystems. All rights reserved.
 * 
 * Updates a directory containing Java objects that uses 
 * draft-ryan-java-schema-00.txt to draft-ryan-java-schema-00.txt (revised).
 * You can update the schema or the Java object entries. After updating
 * the schema, you should use the directory server's administration
 * tool to verify that the changes have been properly applied. The Netscape
 * Directory Server 3.0, for example, ignores the "TOP" attribute. If the
 * schema have not been properly updated, use the administration tool
 * to correct it before updating the entries.
 *
 * You should first turn off schema-checking at the directory server 
 * before running this program.
 *
 * usage:
 * java [-Djava.naming.provider.url=<provider_url>] \
 *     UpdateJavaObjects [-h|-l|-s[n]|-u] [-n<dn>] [-p<passwd>] [-a<auth>] 
 *      
 * -h		Prints the usage message
 * -l		Lists the entries that will be updated.
 * -s[n]	Update schema (-sn means use workaround for Netscape schema bugs).
 * -u		Update entries containing Java object
 * -n <dn> 	Use <dn> as the distinguished name for authentication
 * -p <passwd>	Use <passwd> as the password for authentication
 * -a <auth>	Use <auth> as the authentication mechanism. Default is "simple".
 *
 * If neither -s, -u, -l, -h has been specified, the default is "-l".
 *
 * Here are some examples:
 * The following lists the changes that will be made:
 *     #java UpdateJavaObjects
 *
 * The following updates the schema (using the workaround for Netscape schema
 * bugs), logging in as "cn=directory manager" with the password
 * "secret".
 *     #java UpdateJavaObjects -sn "-ncn=directory manager" -psecret
 *
 * @author Rosanna Lee
 */

import javax.naming.*;
import javax.naming.directory.*;
import java.util.Hashtable;

public class UpdateJavaObjects {
    static final int LIST = 0;
    static final int UPDATE_SCHEMA = 1;
    static final int UPDATE_OBJECTS = 2;

    static String dn, passwd, auth;
    static boolean netscapebug;

    public static void main(String[] args) {

	int command = processCommandLine(args);
	try {
	    DirContext ctx = signOn();
	    switch (command) {
	    case LIST:
		updateObjects(ctx, true);
		break;
	    case UPDATE_SCHEMA:
		updateSchema(ctx);
		break;
	    case UPDATE_OBJECTS:
		updateObjects(ctx, false);
		break;
	    }
	} catch (NamingException e) {
	    e.printStackTrace();
	}
    }

    /**
     * Signs on to directory server using parameters supplied to program.
     * @return The initial context to the server.
     */
    static private DirContext signOn() throws NamingException {
	if (dn != null && auth == null) {
	    auth = "simple"; 	// use simple for Netscape
	}

	Hashtable env = new Hashtable();
	env.put(Context.INITIAL_CONTEXT_FACTORY, 
	    "com.sun.jndi.ldap.LdapCtxFactory");

	if (auth != null) {
	    env.put(Context.SECURITY_AUTHENTICATION, auth);
	    env.put(Context.SECURITY_PRINCIPAL, dn);
	    env.put(Context.SECURITY_CREDENTIALS, passwd);
	}

	// Workaround for Netscape schema bugs
	if (netscapebug) {
	    env.put("com.sun.naming.netscape.schemaBugs", "true");
	}

	return new InitialDirContext(env);
    }

    /**
     * Updates the schema:
     *
     * Delete obsolete attributes:
     * 	javaSerializedObject
     * 	javaFactoryLocation
     * 	javaReferenceAddress
     * 	javaFactory
     * 	javaClassName
     * Add new and updated attributes:
     * 	javaSerializedData
     * 	javaCodebase
     * 	javaClassName
     *  javaRemoteLocation
     *  javaFactory
     *  javaReferenceAddress
     *
     * Delete obsolete object classes:
     *  javaObject
     *  javaNamingReference
     * Add new and updated object classes:
     *  javaObject
     *  javaSerializedObject
     *  javaNamingReference
     *  javaRemoteObject
     */
    static private void updateSchema(DirContext ctx) throws NamingException {

// Attributes
	DirContext attrRoot = (DirContext)
	    ctx.getSchema("").lookup("AttributeDefinition");

	/* Get rid of old attr IDs */
	try {
	    attrRoot.destroySubcontext("javaSerializedObject");
	} catch (NameNotFoundException e) {}
	try {
	    attrRoot.destroySubcontext("javaFactoryLocation");
	} catch (NameNotFoundException e) {}
	try {
	    attrRoot.destroySubcontext("javaReferenceAddress");
	} catch (NameNotFoundException e) {}
	try {
	    attrRoot.destroySubcontext("javaFactory");
	} catch (NameNotFoundException e) {}
	try {
	    attrRoot.destroySubcontext("javaClassName");
	} catch (NameNotFoundException e) {}

	/* Add new and updated attr definitions */
// javaSerializedData
	Attributes attrs = new BasicAttributes(true); // ignore case
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.1.8");
	attrs.put("NAME", "javaSerializedData");
	attrs.put("DESC", "Serialized form of a Java object");
	attrs.put("SYNTAX", "1.3.6.1.4.1.1466.115.121.1.5");
	attrs.put("SINGLE-VALUE", "true");
	attrRoot.createSubcontext("javaSerializedData", attrs);

// javaCodebase
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.1.7");
	attrs.put("NAME", "javaCodebase");
	attrs.put("DESC", "URL specifying the location of class definition");
	attrs.put("EQUALITY", "caseExactIA5Match");
	attrs.put("SYNTAX", "1.3.6.1.4.1.1466.115.121.1.26");
	attrRoot.createSubcontext("javaCodebase", attrs);

// javaClassName
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.1.6");
	attrs.put("NAME", "javaClassName");
	attrs.put("DESC", "Fully qualified Java class or interface name");
	attrs.put("EQUALITY", "caseExactMatch");
	attrs.put("SYNTAX", "1.3.6.1.4.1.1466.115.121.1.15");
	attrs.put("SINGLE-VALUE", "true");
	attrRoot.createSubcontext("javaClassName", attrs);

// javaRemoteLocation
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.1.9");
	attrs.put("NAME", "javaRemoteLocation");
	attrs.put("DESC", "Location of a Java RMI remote object");
	attrs.put("EQUALITY", "caseExactIA5Match");
	attrs.put("SYNTAX", "1.3.6.1.4.1.1466.115.121.1.26");
	attrs.put("SINGLE-VALUE", "true");
	attrRoot.createSubcontext("javaRemoteLocation", attrs);

// javaFactory
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.1.10");
	attrs.put("NAME", "javaFactory");
	attrs.put("DESC", "Fully qualified Java class name of a JNDI object factory");
	attrs.put("EQUALITY", "caseExactMatch");
	attrs.put("SYNTAX", "1.3.6.1.4.1.1466.115.121.1.15");
	attrs.put("SINGLE-VALUE", "true");
	attrRoot.createSubcontext("javaFactory", attrs);

// javaReferenceAddress
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.1.11");
	attrs.put("NAME", "javaReferenceAddress");
	attrs.put("DESC", "Addresses associated with a JNDI Reference");
	attrs.put("EQUALITY", "caseExactMatch");
	attrs.put("SYNTAX", "1.3.6.1.4.1.1466.115.121.1.15");
	attrRoot.createSubcontext("javaReferenceAddress", attrs);

// Object Classes
	DirContext ocRoot = (DirContext)
	    ctx.getSchema("").lookup("ClassDefinition");

	/* Get rid of old OCs */
	try {
	    ocRoot.destroySubcontext("javaObject");
	} catch (NameNotFoundException e) {}
	try {
	    ocRoot.destroySubcontext("javaNamingReference");
	} catch (NameNotFoundException e) {}

// javaObject
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.2.4");
	attrs.put("NAME", "javaObject");
	attrs.put("DESC", "Java object representation");
	attrs.put("SUP", "top");
	attrs.put("ABSTRACT", "true");
	attrs.put("MUST", "javaClassName");
	attrs.put("MAY",  "javaCodebase" );
	ocRoot.createSubcontext("javaObject", attrs);

// javaSerializedObject
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.2.5");
	attrs.put("NAME",  "javaSerializedObject");
	attrs.put("DESC", "Java serialized object");
	attrs.put("SUP", "javaObject");
	attrs.put("AUXILIARY", "true");
	attrs.put("MUST",  "javaSerializedData" );
	ocRoot.createSubcontext("javaSerializedObject", attrs);

// javaRemoteObject
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.2.6");
	attrs.put("NAME", "javaRemoteObject");
	attrs.put("DESC", "Java RMI remote object");
	attrs.put("SUP", "javaObject");
	attrs.put("AUXILIARY", "true");
	attrs.put("MUST", "javaRemoteLocation" );
	ocRoot.createSubcontext("javaRemoteObject", attrs);

// javaNamingReference
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.2.7");
	attrs.put("NAME", "javaNamingReference");
	attrs.put("DESC", "JNDI reference");
	attrs.put("SUP", "javaObject");
	attrs.put("AUXILIARY", "true");
	Attribute attr = new BasicAttribute("MAY", "javaReferenceAddress");
	attr.add("javaFactory");
	attrs.put(attr);
	ocRoot.createSubcontext("javaNamingReference", attrs);

// javaContainer
	attrs = new BasicAttributes(true);
	attrs.put("NUMERICOID", "1.3.6.1.4.1.42.2.27.4.2.1");
	attrs.put("NAME", "javaContainer");
	attrs.put("DESC", "Container for a Java object");
	attrs.put("SUP", "top");
	attrs.put("STRUCTURAL", "true");
	attrs.put("MUST", "cn");
	try {
	    ocRoot.createSubcontext("javaContainer", attrs);
	} catch (NameAlreadyBoundException e) {
	    // may have already existed from previous 00 schema
	}

	System.out.println(
"Please use your directory server's administration tool to verify");
	System.out.println(
"the correctness of the schema before proceeding to update the entries.");
    }

    /**
      * Changes to be performed on entries containing Java objects:
      * For entries containing the javaSerializedObject attribute:
      * o Save attribute value with javaSerializedData attribute
      * o Delete javaSerializedObject attribute
      *
      * For entries containing the javaFactoryLocation attribute:
      * o Save attribute value with javaCodebase attribute
      * o Delete javaFactoryLocation attribute
      *
      * For entries containing "javaObject" in  objectclass attribute:
      * o add "javaSerializedObject" as value
      *
      * For entries containing "javaNamingReference" in  objectclass attribute:
      * o add "javaObject" as value
      */
    static private void updateObjects(DirContext ctx, boolean listOnly) 
	throws NamingException {

	SearchControls ctls = new SearchControls();
	ctls.setSearchScope(SearchControls.SUBTREE_SCOPE);
	ctls.setReturningAttributes(new String[] {
	    "javaSerializedObject", "javaFactoryLocation", "objectclass"});

	// Update serialized objects
	NamingEnumeration enum = ctx.search("", 
	    "(&(javaSerializedObject=*) (objectclass=javaobject))", ctls);

	while (enum.hasMore()) {
	    SearchResult sr = (SearchResult)enum.next();
	    System.out.println(sr.getName());
	    ModificationItem mods[] =
		getModifications(sr.getAttributes(), "javaSerializedObject",
		    "javaSerializedData", "javaSerializedObject");

	    if (listOnly) {
		printMods(mods);
	    } else {
		ctx.modifyAttributes(sr.getName(), mods);
	    }
	}


	// Update references
	enum = ctx.search("", 
	    "(&(objectclass=javanamingreference) (!(objectclass=javaobject)))",
	    ctls);

	while (enum.hasMore()) {
	    SearchResult sr = (SearchResult)enum.next();
	    System.out.println(sr.getName());
	    ModificationItem mods[] = 
		getModifications(sr.getAttributes(), "javaFactoryLocation",
		    "javaCodebase", "javaObject");

	    if (listOnly) {
		printMods(mods);
	    } else {
		ctx.modifyAttributes(sr.getName(), mods);
	    }
	}
    }

    /**
     * Create a list of modifications.
     */
    private static ModificationItem[] getModifications(Attributes attrs,
	String oldAttrID, String newAttrID, String newOC) throws NamingException {
	    Attribute oldAttr = attrs.get(oldAttrID);
	    ModificationItem[] mods;
	    int count = 0;

	    if (oldAttr != null) {
		Attribute newAttr = new BasicAttribute(newAttrID);
		newAttr.add(oldAttr.get());
		mods = new ModificationItem[3];
		mods[count++] = new ModificationItem(
		    DirContext.REMOVE_ATTRIBUTE, oldAttr);
		mods[count++] = new ModificationItem(
		    DirContext.ADD_ATTRIBUTE, newAttr);
	    } else {
		mods = new ModificationItem[1];
	    }

	    mods[count++] = new ModificationItem(
		DirContext.ADD_ATTRIBUTE, new BasicAttribute("objectclass", newOC));

	    return mods;
    }

    private static void printMods(ModificationItem[] mods) {
	for (int i = 0; i < mods.length; i++) {
	    System.out.println("\t" + mods[i]);
	}
    }


    static private int processCommandLine(String[] args) {
	String option;
	boolean schema = false;
	boolean objects = false;
	boolean list = false;

	for (int i = 0; i < args.length; i++) {
	    option = args[i];
	    if (option.startsWith("-h")) {
		printUsage(null);
	    }
	    if (option.startsWith("-s")) {
		schema = true;
		netscapebug = option.equals("-sn");
	    } else if (option.startsWith("-l")) {
		list = true;
	    } else if (option.startsWith("-u")) {
		objects = true;
	    } else if (option.startsWith("-a")) {
		auth = option.substring(2);
	    } else if (option.startsWith("-n")) {
		dn = option.substring(2);
	    } else if (option.startsWith("-p")) {
		passwd = option.substring(2);
	    } else {
		// invalid option
		printUsage("Invalid option");
	    }
	}

	if (list) {
	    if (schema || objects) {
		printUsage("Too many options");
	    }
	} else if (schema && objects) {
	    printUsage("Too many options");
	} else if (!schema && !objects) {
	    return LIST;
	} else if (schema) {
	    return UPDATE_SCHEMA;
	} else {
	    return UPDATE_OBJECTS;
	}
	return LIST;
    }

    static private void printUsage(String msg) {
	if (msg != null) {
	    System.out.println(msg);
	}

System.out.print("Usage: ");
System.out.println("java [-Djava.naming.provider.url=<provider_url>] \\");
System.out.println("  UpdateJavaObjects [-h|-l|-s[n]|-u] [-n<dn>] [-p<passwd>] [-a<auth>]");
System.out.println();
System.out.println("  -h\t\tPrints the usage message");
System.out.println("  -l\t\tPrints the entries that will be updated");
System.out.println("  -s[n]\t\tUpdate schema");
System.out.println("\t\t(-sn means use workaround for Netscape schema bugs)");
System.out.println("  -u\t\tUpdate entries containing Java object");
System.out.println("  -n <dn>\tUse <dn> as the distinguished name for authentication");
System.out.println("  -p <passwd>\tUse <passwd> as the password for authentication");
System.out.println("  -a <auth>\tUse <auth> as the authentication mechanism");
System.out.println("\t\t Default is 'simple' if dn specified; otherwise 'none'");
	System.exit(-1);
    }
}
